/* FILE: display.h                 -*-Mode: c++-*-
 *
 * C++ classes for display windows
 *
 */

#ifndef DISPLAY_H
#define DISPLAY_H

#include "nb.h"
#include "vf.h"

#include "mmdinit.h"

/* End includes */     /* Optional directive to build.tcl */

//////////////////////////////////////////////////////////////////////////
// Colormap class
//
#define TKCOLORLENGTH 6
class DisplayColorMap {
private:
  static const ClassDoc class_doc;
  int colorcount;  // Number of colors in colormap
  Nb_DString mapname;
  char **TkColor;
  void Alloc(int _colorcount);
  void Free();
public:
  void Setup(int _colorcount,const char* _mapname,
	     const char* MakeColor(REAL4m relative_shade));
  DisplayColorMap(void);
  DisplayColorMap(int _colorcount,const char* _mapname,
		  const char* MakeColor(REAL4m relative_shade)) {
    Setup(_colorcount,_mapname,MakeColor);
  }
  DisplayColorMap(const DisplayColorMap&);
  DisplayColorMap& operator=(const DisplayColorMap&);
  ~DisplayColorMap() { Free(); }

  // Setup routines using predefined maps
  void Setup(int _colorcount,const char* _mapname);
  DisplayColorMap(int _colorcount,const char* _mapname) {
    Setup(_colorcount,_mapname);
  }

  int GetColorCount() { return colorcount; }

  const char* GetColor(REAL4m relative_shade) const; // Given a floating
  /// point number between 0. and 1., returns a char* to the closest
  /// corresponding value in the colormap.  The returned pointer will be
  /// valid until the colormap is changed.

  const char* GetColor(INT4m index) const;
  /// Retrieves color based on index.
};

// The predefined MakeColor colormaps
const char* MakeColorRedBlackBlue(REAL4m relative_shade);
const char* MakeColorBlueWhiteRed(REAL4m relative_shade);
const char* MakeColorTealWhiteRed(REAL4m relative_shade);
const char* MakeColorBlackGrayWhite(REAL4m relative_shade);
const char* MakeColorWhiteGreenBlack(REAL4m relative_shade);

//////////////////////////////////////////////////////////////////////////
// PlotConfiguration structure
//
class PlotConfiguration {
private:
  static const ClassDoc class_doc;
public:
  PlotConfiguration();
  PlotConfiguration(const PlotConfiguration &);
  PlotConfiguration& operator=(const PlotConfiguration &);
  
  BOOL displaystate;    // 1=>display, 0=>don't display
  REAL4m subsample;    // Subsample period, in units of mesh cells.
  ///   NOTE: If subsample=0, then *all* points are selected.
  /// Also if subsample<1, then features will be scaled as
  /// though subsample==1.

  REAL4m mag;   // Item magnification, relative to its
  ///  "standard size".  The drawn size depends upon mag,
  /// subsample, and the "standard size".

  Nb_DString colorquantity;

  DisplayColorMap colormap;
  const char* Color(REAL4m shade) const {
    return colormap.GetColor(shade);
  }
};

//////////////////////////////////////////////////////////////////////////
// Display frame class
//
//  Restrictions: Assumes square pixels.
//                2D, ignores z.
//
enum CoordinateSystem { // All are righthanded systems
  CS_ILLEGAL=-1,             // Illegal value
  CS_CALCULATION_STANDARD=0, // x to left,  y up,       z out -  mesh units
  CS_DISPLAY_STANDARD=1,     // x to left,  y down,     z in  - pixel units
  CS_DISPLAY_ROT_90=2,       // x down,     y to right, z in  - pixel units
  CS_DISPLAY_ROT_180=3,      // x to right, y up,       z in  - pixel units
  CS_DISPLAY_ROT_270=4       // x up,       y to left,  z in  - pixel units
};

INT4m CoordinateAngleDifference(CoordinateSystem cs1,CoordinateSystem cs2,
				double &angle);
// Calculates the difference (in degrees, from 0 to 360) of cs1 - cs2, if
// both are CS_DISPLAY_* coordinates, sets the difference in 'angle' and
// returns 0;  If cs1 is not a CS_DISPLAY_* c.s., then the return value
// is 1; if cs2 is not a CS_DISPLAY_* c.s., then the return value is 2.
// In these latter two cases angle is 0.0 on return.

struct SampleInfo {
  REAL4m subsample;
  Nb_DString colorquantity;
  Nb_List<Vf_DisplayVector> dvlist;
  SampleInfo() { subsample=0.; } 
};

class DisplayFrame
{
private:
  static const ClassDoc class_doc;
  static const REAL4m default_arrow_cellsize;
  static const REAL4m default_pixel_cellsize; 
  /// Preferred cellsizes, in pixels.

  static const REAL4m default_arrow_ratio;
  static const REAL4m default_pixel_ratio;
  /// Default (base) size of objects, relative to the cell size.

  Tcl_Interp* tcl_interp;
  Vf_Mesh *mesh;

  void GetBoundingBoxes(Nb_BoundingBox<REAL4> &bdry,
			Nb_BoundingBox<REAL4> &data,REAL8m &data_pad) const;

  INT4m GetDisplayList(SampleInfo &si);

  REAL4m mesh_per_pixel;  // * pixels => mesh units;   This is
  ///  guaranteed>0.  User zoom is given by
  ///    mesh->GetApproximateCellWidth()/mesh_per_pixel;

  REAL4m preferred_arrow_cellsize,preferred_pixel_cellsize;

  CoordinateSystem coords; // Current coordinates
  void ChangeCoordinates(CoordinateSystem old_coords,
			 CoordinateSystem new_coords,
			 Nb_List<Vf_DisplayVector> &display_list) const;
  void ChangeCoordinates(CoordinateSystem old_coords,
			 CoordinateSystem new_coords,
			 Nb_List< Nb_Vec3<REAL4> > &pt_list) const;

  PlotConfiguration arrow_config,pixel_config;

  SampleInfo arrow_sample,pixel_sample;
  /// Current mesh sample list and support information.

  BOOL arrow_valid,pixel_valid,boundary_valid;
  /// If true, then the corresponding mesh sample lists are up-to-date.
  /// In general, display_lists are only updated when a Show request
  /// is made.  Exceptions are scaling and coordinate change requests,
  /// for which valid lists will be adjusted at once.

  void MarkAllSampleListsInvalid() {
    arrow_valid=pixel_valid=boundary_valid=0;
  }

  REAL8m GetPixelSize() const;
  REAL8m GetArrowSize() const;

  INT4m DrawArrow(const char *canvas_name,const Vf_DisplayVector &locvec,
                  REAL4m arrowsize) const;
  INT4m DrawPixel(const char *canvas_name,const Vf_DisplayVector &locvec,
                  REAL4m pixelsize) const;

  BOOL draw_boundary;        // 1=>draw boundary, 0=>don't draw boundary
  Nb_List< Nb_Vec3<REAL4> > boundary_list; // List of boundary vertices
  INT4m DrawLine(const char *canvas_name,
		 const Nb_Vec3<REAL4> &Pt1,const Nb_Vec3<REAL4> &Pt2);

  Nb_DString background_color;

  DisplayFrame(DisplayFrame&);            // Disable copy constructor
  DisplayFrame& operator=(DisplayFrame&); // and assignment operator

public:
  DisplayFrame();
  void SetMesh(Vf_Mesh *newmesh) {
    mesh=newmesh;
    arrow_valid=pixel_valid=boundary_valid=0;
  }
  void SetTclInterp(Tcl_Interp *newinterp) { tcl_interp=newinterp; }

  void GetPlotConfiguration(PlotConfiguration& _arrow_config,
			    PlotConfiguration& _pixel_config) const {
    _arrow_config=arrow_config;
    _pixel_config=pixel_config;
  }
  void SetPlotConfiguration(const PlotConfiguration& _arrow_config,
			    const PlotConfiguration& _pixel_config);

  void SetPreferredArrowCellsize(REAL4m pixels);
  void SetPreferredPixelCellsize(REAL4m pixels);
  REAL4m GetAutoSampleRate(REAL4m cellsize) const;
  /// Returns recommended subsample rate, for the current mesh
  /// at current zoom.

  CoordinateSystem GetCoordinates() const { return coords; }
  BOOL SetCoordinates(CoordinateSystem new_coords);

  const char* GetBackgroundColor() { return background_color; }
  void SetBackgroundColor(const char* color) {
    background_color=color;
  }

  BOOL GetDrawBoundary() const { return draw_boundary; }
  void SetDrawBoundary(BOOL newstate) { draw_boundary=newstate; }

  INT4m GetColorQuantityTypes(Nb_List<Nb_DString> &types) const {
    return mesh->ColorQuantityTypes(types);
  }

  const char* GetVecColor(const Nb_Vec3<REAL4>& v) const;
  /// Access to the underlying Vf_Mesh::GetVecShade function.  See that
  /// routine for details on shade determination.  The mapping from
  /// shade to color depends on the state of arrow_config.displaystate
  /// and pixel_config.displaystate.  Pixel config colorquantity and
  /// colormap are used, unless pixels are turned off and arrows are
  /// turned on, in which case arrrow config colorquantity and colormap
  /// are used.

  INT4m Show(const char *canvas_name);

  REAL4m GetZoom(void) const;
  REAL4m SetZoom(REAL4m zoom);
  REAL4m SetZoom(REAL4m width,REAL4m height);
  /// NOTE: "zoom" is in units of "pixels per approx cell size".

  void GetRequestedSubsampleRates(REAL4m &arrow_samplerate,
				  REAL4m &pixel_samplerate) const {
    arrow_samplerate=arrow_config.subsample;
    pixel_samplerate=pixel_config.subsample;
  }
  void GetActualSubsampleRates(REAL4m &arrow_samplerate,
			       REAL4m &pixel_samplerate) const {
    arrow_samplerate=arrow_sample.subsample;
    pixel_samplerate=pixel_sample.subsample;
  }

  Nb_BoundingBox<REAL4> GetDisplayBox() const;
  /// Returns extents of mesh, in pixels (at current zoom).
};

#endif // DISPLAY_H

